/*
Copyright 2021 The Matrix.org Foundation C.I.C.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

export function createEvent(type, id = null, sender = null) {
    return {type, event_id: id, sender};
}

export function withContent(content, event) {
    return Object.assign({}, event, {content});
}

export function withSender(sender, event) {
    return Object.assign({}, event, {sender});
}

export function withTextBody(body, event) {
    return withContent({body, msgtype: "m.text"}, event);
}

export function withTxnId(txnId, event) {
    return Object.assign({}, event, {unsigned: {transaction_id: txnId}});
}

export function withRedacts(redacts, reason, event) {
    return Object.assign({redacts, content: {reason}}, event);
}

export function withReply(replyToId, event) {
    return withContent({
        "m.relates_to": {
            "m.in_reply_to": {
                "event_id": replyToId
            }
        }
    }, event);
}
